"""
VObject Overview
================
    vobject parses vCard or vCalendar files, returning a tree of Python objects.
    It also provids an API to create vCards or vCalendars data structures which
    can then be serialized.

    Parsing existing streams
    ------------------------
    Streams containing one or many L{Component<vobject.Component>}s can be
    parsed using L{readComponents<vobject.readComponents>}.  As each Component
    is parsed, vobject will attempt to give it a L{Behavior<behavior.Behavior>}.
    If an appropriate Behavior is found, any base64, quoted-printable, or
    backslash escaped data will automatically be decoded.  Dates and datetimes
    will be transformed to datetime.date or datetime.datetime instances.
    Components containing recurrence information will have a special rruleset
    attribute (a dateutil.rrule.rruleset instance).

    Validation
    ----------
    L{Behavior<behavior.Behavior>} classes implement validation for
    L{Component<vobject.Component>}s.  To validate, an object must have all
    required children.  There (TODO: will be) a toggle to raise an exception or
    just log unrecognized, non-experimental children and parameters.
    
    Creating objects programatically
    --------------------------------
    A L{Component<vobject.Component>} can be created from scratch.  No encoding
    is necessary, serialization will encode data automatically.  Factory
    functions (TODO: will be) available to create standard objects.

    Serializing objects
    -------------------
    Serialization: 
      - Looks for missing required children that can be automatically generated,
        like a UID or a PRODID, and adds them
      - Encodes all values that can be automatically encoded
      - Checks to make sure the object is valid (unless this behavior is
        explicitly disabled)
      - Appends the serialized object to a buffer, or fills a new
        buffer and returns it
    
    Examples
    --------
    
    >>> import datetime
    >>> import dateutil.rrule as rrule
    >>> x = iCalendar()
    >>> x.add('vevent')
    <VEVENT (with rruleset)| []>
    >>> x
    <VCALENDAR| [<VEVENT (with rruleset)| []>]>
    >>> v = x.vevent[0]
    >>> utc = vobject.registry.getTzid('UTC')
    >>> v.setDtstart(datetime.datetime(2004, 12, 15, 14, tzinfo = utc))
    >>> v
    <VEVENT (with rruleset)| [<DTSTART{}2004-12-15 14:00:00+00:00>]>
    >>> x
    <VCALENDAR| [<VEVENT (with rruleset)| [<DTSTART{}2004-12-15 14:00:00+00:00>]>]>
    >>> v.rrule.append(rrule.rrule(rrule.WEEKLY, count=2, dtstart=v.getDtstart()))
    >>> list(v.rruleset)
    [datetime.datetime(2004, 12, 15, 14, 0, tzinfo=tzutc()), datetime.datetime(2004, 12, 22, 14, 0, tzinfo=tzutc())]
    >>> v.add('uid').value = "randomuid@MYHOSTNAME"
    >>> print x.serialize().replace(vobject.CRLF, vobject.LF).strip()
    BEGIN:VCALENDAR
    VERSION:2.0
    PRODID:-//PYVOBJECT//NONSGML Version 1//EN
    BEGIN:VEVENT
    UID:randomuid@MYHOSTNAME
    DTSTART:20041215T140000Z
    RDATE:20041222T140000Z
    END:VEVENT
    END:VCALENDAR
    
"""
import vobject, icalendar, StringIO
from vobject import Stack, Component, readComponents

def iCalendar():
    obj = vobject.Component('VCALENDAR')
    obj.setBehavior(icalendar.VCalendar2_0)
    return obj

#----------------------------- Parsing helper ----------------------------------

def readTZ(stream):
    """Register all VTIMEZONE components from stream, restore position."""
    startpos = stream.tell()
    for line, pos, n in vobject.getLogicalLines(stream, False):
        vline = vobject.textLineToContentLine(line)
        if vline.value == 'VTIMEZONE' and vline.name == 'BEGIN':
            recall = stream.tell()
            stream.seek(pos)
            tz=readComponents(stream).next()
            tz.setBehavior(icalendar.VTimezone)
            tz.transformToNative()
            stream.seek(recall)
    stream.seek(startpos)
    

def readICalendar(stream, topLevel=False):
    """Generate one iCalendar subcomponent at a time from a stream.

    To accurately parse an iCalendar stream, its VTIMEZONE components must be
    parsed.  Generally, VTIMEZONEs are serialized near the beginning, but this
    isn't required by RFC2445.  To avoid holding a large icalendar stream in
    memory, first scan for VTIMEZONEs using readTZ, then iterate over
    readICalendar().
        
    By default, readICalendar iterates over subcomponents, top level content
    lines are ignored.  If the topLevel flag is set, readICalendar returns a
    VCALENDAR object with no subcomponents, only content line children.
    
    """
    stack = Stack()
    fragmentBuffer = StringIO.StringIO()
    readTZ(stream)
    for line, pos, n in vobject.getLogicalLines(stream, False):
        if not topLevel and len(stack) >= 2:
            fragmentBuffer.write(line)
            fragmentBuffer.write('\r\n')
        vline = vobject.textLineToContentLine(line)
        if   vline.name == "VERSION":
            stack.top().setBehaviorFromVersionLine(vline)
        elif vline.name == "BEGIN":
            stack.push(Component(vline.value))
            if not topLevel and len(stack) == 2:               
                startPos = pos
                fragmentBuffer.write(unicode(line))
                fragmentBuffer.write('\r\n')
        elif vline.name == "END":
            if vline.value == stack.topName():#START matches END
                if len(stack) == 1:
                    if topLevel: yield stack.pop()
                    else: stack.pop()
                elif len(stack) == 2:
                    component = stack.pop()
                    if topLevel: continue
                    try:
                        component.parentBehavior = stack.top().behavior
                        component.autoBehavior(True)
                        component.transformChildrenToNative()
                        component = component.transformToNative()
                        component.fragmentString = fragmentBuffer.getvalue()
                        fragmentBuffer.truncate(0)
                        yield component #EXIT POINT
                    except vobject.VObjectError, e:
                        e.linenumber = n
                        raise
                        
                else: stack.modifyTop(stack.pop())
            else:
                print repr(stack.topName())
                err = "At line %i, %s component wasn't closed"
                raise vobject.VObjectError(err % (n, stack.topName()))
        elif not topLevel or len(stack) == 1:
            stack.modifyTop(vline) #not a START or END line